# Copyright (C) 2023, Advanced Micro Devices, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause


from .weighted_bit_width import WeightBitWidthWeightedBySize, ActivationBitWidthWeightedBySize
from .weighted_bit_width import QuantLayerOutputBitWidthWeightedByOps, MEGA
